; ########################################
; NHWAVEGEN
; WAVEFORMGENERATOR
; (C) BY NILS HINTZE
; ########################################



; ########################################
;- Konstanten
; ########################################

#WAVE_FORMAT_EXTENSIBLE = $FFFE
#CALLBACK_THREAD = #CALLBACK_TASK
#SPEAKER_FRONT_LEFT            = $00001
#SPEAKER_FRONT_RIGHT           = $00002
#SPEAKER_FRONT_CENTER          = $00004
#SPEAKER_LOW_FREQUENCY         = $00008
#SPEAKER_BACK_LEFT             = $00010
#SPEAKER_BACK_RIGHT            = $00020
#SPEAKER_FRONT_LEFT_OF_CENTER  = $00040
#SPEAKER_FRONT_RIGHT_OF_CENTER = $00080
#SPEAKER_BACK_CENTER           = $00100
#SPEAKER_SIDE_LEFT             = $00200
#SPEAKER_SIDE_RIGHT            = $00400
#SPEAKER_TOP_CENTER            = $00800
#SPEAKER_TOP_FRONT_LEFT        = $01000
#SPEAKER_TOP_FRONT_CENTER      = $02000
#SPEAKER_TOP_FRONT_RIGHT       = $04000
#SPEAKER_TOP_BACK_LEFT         = $08000
#SPEAKER_TOP_BACK_CENTER       = $10000
#SPEAKER_TOP_BACK_RIGHT        = $20000

; Gre der Kanalmaske in Bits (Anzahl von #SPEAKER-Konstanten)
#WaveformGenerator_ChannelMaskBitSize = 18

; Maximale Kanalanzahl
#WaveformGenerator_MaximumChannelCount = 8

; Messages fr Thread
Enumeration #WM_APP
  #WG_WM_Start
  #WG_WM_Stop
EndEnumeration

; Wellenformen
Enumeration
  #WG_Sine
  #WG_Triangle
  #WG_SawtoothFalling
  #WG_SawtoothRising
  #WG_Rectangle
  #WG_Arbitrary
EndEnumeration

; WaveformGenerator\Flags
#WG_Enable = $01 ; WaveformGenerator aktiviert

; WG_ChannelSettings\Flags
#WG_Mute = $01 ; Stummschaltung aktiviert
#WG_FMEnable = $02 ; Frequenzmodulation aktiviert
#WG_AMEnable = $04 ; Amplitudenmodulation aktiviert

; Arbitrrdaten
Enumeration
  #WG_ArbitraryData_Waveform
  #WG_ArbitraryData_FMWaveform
  #WG_ArbitraryData_AMWaveform
  #WG_ArbitraryDataCount
EndEnumeration

; Workaround fr fehlende GetThreadId_()-Funktion vor Windows Vista bzw. Windows Server 2003
#WaveformGenerator_WindowsXPWorkaround = #True



; ########################################
;- Strukturen
; ########################################

Structure WG_Pointer
  *p
EndStructure

Structure WG_ChannelSettings
  Flags.i ; Diverse Flags
  Waveform.i ; Wellenform
  PhaseShift.d ; Phasenverschiebung [rad]
  Frequency.d ; Frequenz [Hz]
  FMWaveform.i ; FM-Wellenform
  FMFrequency.d ; FM-Frequenz [Hz]
  FMMinFreq.d ; Minimale FM-Frequenz [Hz]
  FMMaxFreq.d ; Maximale FM-Frequenz [Hz]
  Amplitude.d ; Amplitude [1]
  AMWaveform.i ; AM-Wellenform
  AMFrequency.d ; AM-Frequenz [Hz]
  AMMinAmp.d ; Minimale AM-Amplitude [1]
  AMMaxAmp.d ; Maximale AM-Amplitude [1]
EndStructure

Structure WG_ArbitraryData
  *Points ; Arbitrrdaten, Zeiger auf Array von Double-Werten, muss auf einen gltigen Speicherbereich zeigen oder 0 sein
  PointCount.i ; Punktanzahl der Arbitrrdaten, muss korrekt sein wenn \Points auf einen gltigen Speicherbereich zeigt
EndStructure

Structure WG_Channel
  Settings.WG_ChannelSettings ; Kanaleinstellungs-Struktur, Zugriff nur im Hauptprogramm
  ; Intern
  NewSettings.WG_ChannelSettings ; Kanaleinstellungs-Struktur, geteilt zwischen Hauptprogramm und Thread, geschtzt durch Mutex
  CurrentSettings.WG_ChannelSettings ; Kanaleinstellungs-Struktur, Zugriff nur im Thread
  NewArbitraryData.WG_ArbitraryData[#WG_ArbitraryDataCount] ; Neue Arbitrrdaten zur bernahme durch den Thread, geschtzt durch Mutex
  CurrentArbitraryData.WG_ArbitraryData[#WG_ArbitraryDataCount] ; Aktuelle Arbitrrdaten, Zugriff nur im Thread
EndStructure

Structure WaveformGenerator
  DeviceID.i ; ID des Ausgabegertes
  SampleRate.i ; Abtastrate [Hz]
  BitsPerSample.i ; Auflsung [Bit]
  BufferLength.i ; Pufferlnge [ms]
  ChannelMask.i ; Aktive Kanle, Kombination von #SPEAKER-Konstanten
  ChannelCount.i ; Anzahl aktiver Kanle = Anzahl der gesetzten Bits in \ChannelMask
  Channel.WG_Channel[#WaveformGenerator_MaximumChannelCount]
  ; Intern
  Flags.i ; Diverse Flags
  Ticks.q ; Vergangene Zeit seit dem Start [Samples]
  Mutex.i ; Mutex zum geteilten Zugriff auf Variablen durch Hauptprogramm und Thread
  Semaphore.i ; 
  Thread.i ; PureBasic Thread-Nummer
  ThreadID.i ; WinAPI Thread-ID
  *hWaveOut ; Handle des geffneten Ausgabegertes
  BufferSize.i ; Puffergre [Samples]
  Buffer1.WAVEHDR ; Header des ersten Ausgabepuffers
  Buffer2.WAVEHDR ; Header des zweiten Ausgabepuffers
EndStructure



; ########################################
;- Prozeduren
; ########################################

; Thread-ID anhand des Handles ermitteln
CompilerIf Not #WaveformGenerator_WindowsXPWorkaround
  Import "Kernel32.lib"
    GetThreadId.i(*Handle)
  EndImport
CompilerEndIf

; WaveOut-Fehlertext ermitteln
Procedure.s WG_WaveOutErrorText(MMResult)
  Error.s{255}
  waveOutGetErrorText_(MMResult, @Error, 255)
  ProcedureReturn PeekS(@Error)
EndProcedure

; Schreibt ein Sample an die angegebene Adresse und inkrementiert den Zeiger
; Value muss zwischen -1 und +1 liegen.
; *Buffer ist ein ZEIGER AUF DEN ZEIGER, welcher die Adresse enthlt.
; BitsPerSample muss 8, 16, 24 oder 32 sein.
Procedure WG_WriteSample(Value.d, *Pointer.WG_Pointer, BitsPerSample)
  If Value > 1
    Value = 1
  ElseIf Value < -1
    Value = -1
  EndIf
  BitsPerSample >> 3 ; BitsPerSample ist jetzt BytesPerSample
  Select BitsPerSample
    Case 1 ; Bei 8 Bit besteht der Puffer aus vorzeichenlosen Bytes
      PokeA(*Pointer\p, $FF*(Value+1)/2)
    Case 2 ; Bei 16 Bit und mehr sind die Werte vorzeichenbehaftet
      PokeW(*Pointer\p, $7FFF*Value)
    Case 3
      CompilerIf #PB_Compiler_Processor <> #PB_Processor_x86
        CompilerError "Dieser Code ist fr eine Little-Endian-Byte-Reihenfolge vorgesehen! Gegebenenfalls anpassen!"
      CompilerEndIf
      Long.l = $7FFFFFFF * Value
      CopyMemory(@Long+1, *Pointer\p, 3) ; @Long + 1 = @Long + SizeOf(Long) - BitsPerSample
    Case 4
      PokeL(*Pointer\p, $7FFFFFFF*Value)
  EndSelect
  *Pointer\p + BitsPerSample
EndProcedure

; Dreieckfunktion
Procedure.d WG_Triangle(Angle.d)
  Angle + (#PI / 2)
  Angle = Mod(Angle, 2*#PI)
  If Angle < 0
    Angle + (2 * #PI)
  EndIf
  If Angle >= #PI
    Angle = 2 * #PI - Angle
  EndIf
  ProcedureReturn Angle / (#PI / 2) - 1
EndProcedure

; Sgezahnfunktion fallend
Procedure.d WG_SawtoothFalling(Angle.d)
  Angle = Mod(Angle, 2*#PI)
  If Angle < 0
    Angle + (2 * #PI)
  EndIf
  ProcedureReturn 1 - Angle / #PI
EndProcedure

; Sgezahnfunktion steigend
Procedure.d WG_SawtoothRising(Angle.d)
  Angle = Mod(Angle, 2*#PI)
  If Angle < 0
    Angle + (2 * #PI)
  EndIf
  ProcedureReturn Angle / #PI - 1
EndProcedure

; Rechteckfunktion
Procedure.d WG_Rectangle(Angle.d)
  Angle = Mod(Angle, 2*#PI)
  If Angle < 0
    Angle + (2 * #PI)
  EndIf
  If Angle >= #PI
    ProcedureReturn -1
  EndIf
  ProcedureReturn 1
EndProcedure

; Arbitrrfunktion
Procedure.d WG_ArbitraryFunction(Angle.d, *AD.WG_ArbitraryData)
  With *AD
    If \Points = 0
      ProcedureReturn 0
    EndIf
    Angle = Mod(Angle, 2*#PI)
    If Angle < 0
      Angle + (2 * #PI)
    EndIf
    Angle * (\PointCount / (2 * #PI)) ; Angle ist jetzt reeller Array-Index
    Index = Int(Angle)
    *Point1.Double = \Points + Index * SizeOf(Double)
    *Point2.Double = \Points + ((Index + 1) % \PointCount) * SizeOf(Double)
    Angle = Mod(Angle, 1)
    ProcedureReturn *Point1\d * (1 - Angle) + *Point2\d * Angle
  EndWith
EndProcedure

; 
Procedure.d WG_SinFM(Angle.d)
  ProcedureReturn 1 - Cos(Angle)
EndProcedure

; 
Procedure.d WG_TriangleFM(Angle.d)
  Angle = Mod(Angle, 2*#PI)
  If Angle < #PI / 2
    Angle + (2 * #PI)
  EndIf
  If Angle < #PI * 3 / 2
    Angle = 2 * Angle - Pow(Angle, 2) / #PI - #PI / 2
  Else
    Angle = 4 * #PI + Pow(Angle, 2) / #PI - 4 * Angle
  EndIf
  ProcedureReturn Angle
EndProcedure

; 
Procedure.d WG_SawtoothFallingFM(Angle.d)
  Angle = Mod(Angle, 2*#PI)
  If Angle < 0
    Angle + (2 * #PI)
  EndIf
  ProcedureReturn Angle - Pow(Angle, 2) / (2 * #PI)
EndProcedure

; 
Procedure.d WG_SawtoothRisingFM(Angle.d)
  Angle = Mod(Angle, 2*#PI)
  If Angle < 0
    Angle + (2 * #PI)
  EndIf
  ProcedureReturn Pow(Angle, 2) / (2 * #PI) - Angle
EndProcedure

; 
Procedure.d WG_RectangleFM(Angle.d)
  Angle = Mod(Angle, 2*#PI)
  If Angle < 0
    Angle + (2 * #PI)
  EndIf
  If Angle >= #PI
    Angle = 2 * #PI - Angle
  EndIf
  ProcedureReturn Angle
EndProcedure

; Ausgabepuffer befllen
Procedure WG_FillBuffer(*WG.WaveformGenerator, *Buffer)
  
  With *WG
    
    ; Mutex sperren
    LockMutex(\Mutex)
    
    For n = 0 To \ChannelCount - 1
      
      ; Kanaleinstellungen kopieren
      CopyStructure(@\Channel[n]\NewSettings, @\Channel[n]\CurrentSettings, WG_ChannelSettings)
      
      ; Arbitrrdaten bernehmen
      For i = 0 To #WG_ArbitraryDataCount - 1
        If \Channel[n]\NewArbitraryData[i]\Points
          If \Channel[n]\CurrentArbitraryData[i]\Points
            FreeMemory(\Channel[n]\CurrentArbitraryData[i]\Points)
          EndIf
          CopyStructure(@\Channel[n]\NewArbitraryData[i], \Channel[n]\CurrentArbitraryData[i], WG_ArbitraryData)
          \Channel[n]\NewArbitraryData[i]\Points = 0
          \Channel[n]\NewArbitraryData[i]\PointCount = 0
        EndIf
      Next
        
    Next
    
    ; Mutex entsperren
    UnlockMutex(\Mutex)
    
  EndWith
  
  For n = 0 To *WG\BufferSize - 1
    
    Time.d = *WG\Ticks / *WG\SampleRate
    
    For i = 0 To *WG\ChannelCount - 1
      
      *WGC.WG_Channel = @*WG\Channel[i]
      
      With *WGC\CurrentSettings
        
        If Not \Flags & #WG_Mute
          
          ; Frequenzmodulation
          If \Flags & #WG_FMEnable
            Angle.d = 2 * #PI * Time * (\FMMaxFreq + \FMMinFreq) / 2 + \PhaseShift
            If \FMFrequency <> 0
              Select \FMWaveform
                Case #WG_Sine
                  FM.d = WG_SinFM(2*#PI*\FMFrequency*Time) / (2 * #PI * \FMFrequency)
                Case #WG_Triangle
                  FM.d = WG_TriangleFM(2*#PI*\FMFrequency*Time) / (2 * #PI * \FMFrequency)
                Case #WG_SawtoothFalling
                  FM.d = WG_SawtoothFallingFM(2*#PI*\FMFrequency*Time) / (2 * #PI * \FMFrequency)
                Case #WG_SawtoothRising
                  FM.d = WG_SawtoothRisingFM(2*#PI*\FMFrequency*Time) / (2 * #PI * \FMFrequency)
                Case #WG_Rectangle
                  FM.d = WG_RectangleFM(2*#PI*\FMFrequency*Time) / (2 * #PI * \FMFrequency)
                Case #WG_Arbitrary
                  FM.d = 0
              EndSelect
              Angle = Angle + 2 * #PI * FM * (\FMMaxFreq - \FMMinFreq) / 2
            EndIf
          Else
            Angle.d = 2 * #PI * \Frequency * Time + \PhaseShift
          EndIf
          
          ; Amplitudenmodulation
          If \Flags & #WG_AMEnable
            Select \AMWaveform
              Case #WG_Sine
                Amplitude.d = Sin(2*#PI*\AMFrequency*Time)
              Case #WG_Triangle
                Amplitude.d = WG_Triangle(2*#PI*\AMFrequency*Time)
              Case #WG_SawtoothFalling
                Amplitude.d = WG_SawtoothFalling(2*#PI*\AMFrequency*Time)
              Case #WG_SawtoothRising
                Amplitude.d = WG_SawtoothRising(2*#PI*\AMFrequency*Time)
              Case #WG_Rectangle
                Amplitude.d = WG_Rectangle(2*#PI*\AMFrequency*Time)
              Case #WG_Arbitrary
                Amplitude.d = WG_ArbitraryFunction(2*#PI*\AMFrequency*Time, @*WGC\CurrentArbitraryData[#WG_ArbitraryData_AMWaveform])
            EndSelect
            Amplitude = Amplitude * (\AMMaxAmp - \AMMinAmp) / 2 + (\AMMaxAmp + \AMMinAmp) / 2
          Else
            Amplitude.d = \Amplitude
          EndIf
          
          ; Funktion
          Select \Waveform
            Case #WG_Sine
              Value.d = Sin(Angle) * Amplitude
            Case #WG_Triangle
              Value.d = WG_Triangle(Angle) * Amplitude
            Case #WG_SawtoothFalling
              Value.d = WG_SawtoothFalling(Angle) * Amplitude
            Case #WG_SawtoothRising
              Value.d = WG_SawtoothRising(Angle) * Amplitude
            Case #WG_Rectangle
              Value.d = WG_Rectangle(Angle) * Amplitude
            Case #WG_Arbitrary
              Value.d = WG_ArbitraryFunction(Angle, @*WGC\CurrentArbitraryData[#WG_ArbitraryData_Waveform]) * Amplitude
          EndSelect

        Else ; Stummschaltung
          
          Value.d = 0
          
        EndIf
        
        ; Sample ausgeben
        WG_WriteSample(Value, @*Buffer, *WG\BitsPerSample)
        
      EndWith
      
    Next
    
    *WG\Ticks + 1
    
  Next
  
EndProcedure

; Thread zum Neubefllen und Ausgeben der Ausgabepuffer
Procedure WG_Thread(*WG.WaveformGenerator)
  
  Protected ProcessDoneMessage = #False
  
  Debug "WG_Thread gestartet", 10
  
  CompilerIf #WaveformGenerator_WindowsXPWorkaround
    *WG\ThreadID = GetCurrentThreadId_()
    Debug "WG_Thread hat ThreadID gesetzt", 10
    SignalSemaphore(*WG\Semaphore)
  CompilerEndIf
  
  While GetMessage_(@M.MSG, #Null, 0, 0) > 0
    
    Select M\message
        
      Case #WG_WM_Start
        
        Debug "WG_Thread hat WG_WM_Start empfangen", 10
        
        ; Ausgabepuffer erstmals befllen
        WG_FillBuffer(*WG, *WG\Buffer1\lpData)
        WG_FillBuffer(*WG, *WG\Buffer2\lpData)
        
        ; Ausgabepuffer erstmals ausgeben
        waveOutWrite_(*WG\hWaveOut, @*WG\Buffer1, SizeOf(WAVEHDR))
        waveOutWrite_(*WG\hWaveOut, @*WG\Buffer2, SizeOf(WAVEHDR))
        
        ; MM_WOM_DONE ab jetzt bearbeiten
        ProcessDoneMessage = #True
        
      Case #MM_WOM_DONE
        
        If ProcessDoneMessage
          
          Debug "WG_Thread hat MM_WOM_DONE empfangen", 20
        
          *WH.WAVEHDR = M\lParam
      
          CompilerIf #PB_Compiler_Debugger
            DisableDebugger
            QueryPerformanceFrequency_(@PerfCntFreq.q)
            QueryPerformanceCounter_(@PerfCntStart.q)
          CompilerEndIf
          
          ; Ausgabepuffer neu befllen
          WG_FillBuffer(*WG, *WH\lpData)
      
          CompilerIf #PB_Compiler_Debugger
            QueryPerformanceCounter_(@PerfCntEnd.q)
            EnableDebugger
            Debug "WG_FillBuffer bentigte "+StrD(1000.0*(PerfCntEnd-PerfCntStart)/PerfCntFreq)+" ms", 20
          CompilerEndIf
          
          ; Ausgabepuffer ausgeben
          waveOutWrite_(*WG\hWaveOut, *WH, SizeOf(WAVEHDR))
          
        EndIf
        
      Case #WG_WM_Stop
        
        Debug "WG_Thread hat WG_WM_Stop empfangen", 10
        
        ; Wiedergabe der Ausgabepuffer abbrechen
        waveOutReset_(*WG\hWaveOut)
        
        ; Ab jetzt MM_WOM_DONE ignorieren, damit keine Puffer mehr ausgegeben werden
        ProcessDoneMessage = #False
        
        ; Anzeigen, dass Message bearbeitet wurde
        SignalSemaphore(*WG\Semaphore)
      
    EndSelect
    
  Wend
  
  Debug "WG_Thread beendet", 10
  
EndProcedure

; WaveformGenerator anhalten
Procedure WaveformGeneratorStop(*WG.WaveformGenerator)
  
  With *WG
    
    ; Ausgabe beenden
    If \hWaveOut
      PostThreadMessage_(\ThreadID, #WG_WM_Stop, 0, 0)
      WaitSemaphore(\Semaphore)
    EndIf
    
    ; Header unvorbereiten
    If \hWaveOut
      MMResult = waveOutUnprepareHeader_(\hWaveOut, @\Buffer1, SizeOf(WAVEHDR))
      If MMResult <> #MMSYSERR_NOERROR
        Goto WaveformGeneratorStop_Error
      EndIf
      MMResult = waveOutUnprepareHeader_(\hWaveOut, @\Buffer2, SizeOf(WAVEHDR))
      If MMResult <> #MMSYSERR_NOERROR
        Goto WaveformGeneratorStop_Error
      EndIf
    EndIf
    
    ; Ausgabepuffer freigeben
    If \Buffer1\lpData
      FreeMemory(\Buffer1\lpData)
      \Buffer1\lpData = 0
    EndIf
    If \Buffer2\lpData
      FreeMemory(\Buffer2\lpData)
      \Buffer2\lpData = 0
    EndIf
    
    ; Ausgabegert schlieen
    If \hWaveOut
      MMResult = waveOutClose_(\hWaveOut)
      If MMResult <> #MMSYSERR_NOERROR
        Goto WaveformGeneratorStop_Error
      EndIf
      \hWaveOut = 0
    EndIf
      
    ; Thread beenden
    If \Thread
      PostThreadMessage_(\ThreadID, #WM_QUIT, 0, 0)
      If Not WaitThread(\Thread, 5000)
        Goto WaveformGeneratorStop_Error
      EndIf
      \Thread = 0
    EndIf
    
    ; Semaphore freigeben
    If \Semaphore
      FreeSemaphore(\Semaphore)
      \Semaphore = 0
    EndIf
    
    ; Mutex freigeben
    If \Mutex
      FreeMutex(\Mutex)
      \Mutex = 0
    EndIf
    
    ; Arbitrrdaten freigeben
    For n = 0 To \ChannelCount - 1
      For i = 0 To #WG_ArbitraryDataCount - 1
        If \Channel[n]\NewArbitraryData[i]\Points
          FreeMemory(\Channel[n]\NewArbitraryData[i]\Points)
          \Channel[n]\NewArbitraryData[i]\Points = 0
          \Channel[n]\NewArbitraryData[i]\PointCount = 0
        EndIf
        If \Channel[n]\CurrentArbitraryData[i]\Points
          FreeMemory(\Channel[n]\CurrentArbitraryData[i]\Points)
          \Channel[n]\CurrentArbitraryData[i]\Points = 0
          \Channel[n]\CurrentArbitraryData[i]\PointCount = 0
        EndIf
      Next
    Next
    
    \Flags & ~#WG_Enable
    
    ProcedureReturn  #True
    
WaveformGeneratorStop_Error:

    Debug "WaveformGeneratorStop fehlgeschlagen: "+WG_WaveOutErrorText(MMResult)

    ProcedureReturn #False
    
  EndWith
    
EndProcedure

; WaveformGenerator starten
Procedure WaveformGeneratorStart(*WG.WaveformGenerator)
  
  With *WG
    
    If \Flags & #WG_Enable
      ProcedureReturn #False
    EndIf
    
    ; Zeit zurcksetzen
    \Ticks = 0
    
    ; Mutex erstellen
    \Mutex = CreateMutex()
    If Not \Mutex
      Goto WaveformGeneratorStart_Error
    EndIf
    
    ; Semaphore erstellen
    \Semaphore = CreateSemaphore()
    If Not \Semaphore
      Goto WaveformGeneratorStart_Error
    EndIf
    
    ; Thread erstellen
    \Thread = CreateThread(@WG_Thread(), *WG)
    If Not \Thread
      Goto WaveformGeneratorStart_Error
    EndIf
    CompilerIf Not #WaveformGenerator_WindowsXPWorkaround
      \ThreadID = GetThreadId(ThreadID(\Thread))
    CompilerElse
      Debug "WaveformGeneratorStart wartet, bis WG_Thread ThreadID setzt", 10
      WaitSemaphore(\Semaphore)
    CompilerEndIf
    
    ; Ausgabegert ffnen
    WFE.WAVEFORMATEXTENSIBLE\Format\wFormatTag = #WAVE_FORMAT_EXTENSIBLE
    WFE\Format\nChannels = \ChannelCount
    WFE\Format\nSamplesPerSec = \SampleRate
    WFE\Format\wBitsPerSample = \BitsPerSample
    WFE\Format\nBlockAlign = WFE\Format\nChannels * WFE\Format\wBitsPerSample >> 3
    WFE\Format\nAvgBytesPerSec = WFE\Format\nSamplesPerSec * WFE\Format\nBlockAlign
    WFE\Format\cbSize = 22
    WFE\wValidBitsPerSample = \BitsPerSample
    WFE\dwChannelMask = \ChannelMask
    CopyStructure(?KSDATAFORMAT_SUBTYPE_PCM, @WFE\SubFormat, GUID)
    MMResult = waveOutOpen_(@\hWaveOut, \DeviceID, @WFE, \ThreadID, #Null, #WAVE_FORMAT_DIRECT|#CALLBACK_THREAD)
    If MMResult <> #MMSYSERR_NOERROR
      Goto WaveformGeneratorStart_Error
    EndIf
    
    ; Puffergre berechnen
    \BufferSize = \SampleRate * \BufferLength / 1000
    
    ; Ausgabepuffer reservieren
    \Buffer1\dwBufferLength = \BufferSize * \ChannelCount * \BitsPerSample >> 3
    \Buffer1\lpData = AllocateMemory(\Buffer1\dwBufferLength)
    If Not \Buffer1\lpData
      Goto WaveformGeneratorStart_Error
    EndIf
    \Buffer2\dwBufferLength = \BufferSize * \ChannelCount * \BitsPerSample >> 3
    \Buffer2\lpData = AllocateMemory(\Buffer2\dwBufferLength)
    If Not \Buffer2\lpData
      Goto WaveformGeneratorStart_Error
    EndIf
    
    ; Header vorbereiten
    \Buffer1\dwFlags = 0
    MMResult = waveOutPrepareHeader_(\hWaveOut, @\Buffer1, SizeOf(WAVEHDR))
    If MMResult <> #MMSYSERR_NOERROR
      Goto WaveformGeneratorStart_Error
    EndIf
    \Buffer2\dwFlags = 0
    MMResult = waveOutPrepareHeader_(\hWaveOut, @\Buffer2, SizeOf(WAVEHDR))
    If MMResult <> #MMSYSERR_NOERROR
      Goto WaveformGeneratorStart_Error
    EndIf

    ; Ausgabe starten
    PostThreadMessage_(\ThreadID, #WG_WM_Start, 0, 0)
    
    \Flags | #WG_Enable

    ProcedureReturn #True
    
WaveformGeneratorStart_Error:

    Debug "WaveformGeneratorStart fehlgeschlagen: "+WG_WaveOutErrorText(MMResult)
    
    WaveformGeneratorStop(*WG)

    ProcedureReturn #False

  EndWith
  
EndProcedure

; Prfen, ob WaveformGenerator aktiv ist
Procedure WaveformGeneratorEnabled(*WG.WaveformGenerator)
  If *WG\Flags & #WG_Enable
    ProcedureReturn #True
  EndIf
  ProcedureReturn #False
EndProcedure

; Kanaleinstellungen bernehmen
; Diese Funktion muss zur bernahme der Einstellungen in WaveformGenerator\Channel[n]\Settings aufgerufen werden.
; Die Funktion darf auch aufgerufen werden, wenn der WaveformGenerator nicht aktiv ist.
; Damit knnen die Kanaleinstellungen vor dem Start festgelegt werden.
Procedure WaveformGeneratorApplyChannelSettings(*WG.WaveformGenerator, Channel = -1)
  
  With *WG
    
    ; Mutex sperren
    If \Mutex
      
      CompilerIf #PB_Compiler_Debugger
        QueryPerformanceFrequency_(@PerfCntFreq.q)
        QueryPerformanceCounter_(@PerfCntStart.q)
      CompilerEndIf
      
      LockMutex(\Mutex)
      
      CompilerIf #PB_Compiler_Debugger
        QueryPerformanceCounter_(@PerfCntEnd.q)
        Debug "WaveformGeneratorApplyChannelSettings wartete "+StrD(1000.0*(PerfCntEnd-PerfCntStart)/PerfCntFreq)+" ms auf den Mutex", 20
      CompilerEndIf
      
    EndIf
    
    ; Kanaleinstellungen kopieren
    If Channel = -1
      For n = 0 To \ChannelCount - 1
        CopyStructure(@\Channel[n]\Settings, @\Channel[n]\NewSettings, WG_ChannelSettings)
      Next
    Else
      CopyStructure(@\Channel[Channel]\Settings, @\Channel[Channel]\NewSettings, WG_ChannelSettings)
    EndIf
    
    ; Mutex wieder freigeben
    If \Mutex
      UnlockMutex(\Mutex)
    EndIf
    
  EndWith
  
EndProcedure

; Arbitrrdaten aus Datei lesen
Procedure WaveformGeneratorReadArbitraryFile(*WG.WaveformGenerator, Channel, ArbitraryData, FileName$)
  File = ReadFile(#PB_Any, FileName$)
  If Not File
    ProcedureReturn #False
  EndIf
  NewList Points.d()
  While Not Eof(File)
    Line$ = Trim(ReadString(File))
    If Line$ <> ""
      If AddElement(Points())
        Points() = ValD(Line$)
      EndIf
    EndIf
  Wend
  CloseFile(File)
  PointCount = ListSize(Points())
  If PointCount = 0
    ProcedureReturn #False
  EndIf
  *Points = AllocateMemory(PointCount*SizeOf(Double))
  If Not *Points
    ProcedureReturn #False
  EndIf
  *Point.Double = *Points
  ForEach Points()
    *Point\d = Points()
    *Point + SizeOf(Double)
  Next
  LockMutex(*WG\Mutex)
  If *WG\Channel[Channel]\NewArbitraryData[ArbitraryData]\Points
    FreeMemory(*WG\Channel[Channel]\NewArbitraryData[ArbitraryData]\Points)
  EndIf
  *WG\Channel[Channel]\NewArbitraryData[ArbitraryData]\Points = *Points
  *WG\Channel[Channel]\NewArbitraryData[ArbitraryData]\PointCount = PointCount
  UnlockMutex(*WG\Mutex)
  ProcedureReturn #True
EndProcedure



; ########################################
;- Binrdaten
; ########################################

DataSection
  ; GUIDs
  KSDATAFORMAT_SUBTYPE_PCM:
  Data.l $00000001
  Data.w $0000, $0010
  Data.b $80, $00, $00, $AA, $00, $38, $9B, $71
EndDataSection



; IDE Options = PureBasic 5.21 LTS Beta 3 (Windows - x86)
; Folding = AAA+
; EnableXP
; UseMainFile = nhwavegen.pb